from typing import List, Dict, Optional

from bobtail.route import Route


class Parser:
    routes: List[Route]

    # the incoming request path
    path: str

    meta_data: Dict = None

    def __init__(self, routes: List[Route], path):
        self.path = path
        self.routes = routes

    def _match(self) -> Dict:
        """
        Remove all variable segments from both stings & match
        """
        # Create a fresh metadata dict
        self.meta_data = {"path": {}, "routes": {}, "matched": None}
        # Cache path
        path_segments = self.path.split("/")
        self.meta_data["path"] = {
            "split": path_segments[1:],
        }
        # Cache route path
        for rc, route in self.routes:
            route_class = rc.__class__.__name__
            route_split = route.split("/")
            # match route against path
            self.meta_data["routes"][route_class] = {
                "route": route,
                "split": route_split[1:],
                "vars": None,
            }

        for k, _ in self.meta_data["routes"].items():
            split_route_vals = self.meta_data["routes"][k]["split"]
            split_path_vals = self.meta_data["path"]["split"]
            # check if the incoming request path is longer than the stored route path
            if len(split_route_vals) != len(split_path_vals):
                break
            # route_segment - the assigned route handlers path
            _route_vars = {}
            for i, route_segment in enumerate(split_route_vals):
                # path_segment - the incoming requests path
                path_segment = self.meta_data["path"]["split"][i]
                # Check if path is "/"
                if len(route_segment) == 0 and self.path == "/":
                    self.meta_data["matched"] = k
                    self._set_metadata(k, path_segment, None, None)
                    return self.meta_data["routes"][k]
                # Test route matches path
                if route_segment[0] != "{" and route_segment != path_segment:
                    # No match, break out of this route class
                    break
                if route_segment[0] == "{":
                    # If we reach this point then store the vars
                    n, t = route_segment[1:-1].split(":")
                    self._set_metadata(k, path_segment, t, n)
                if (len(split_path_vals) - 1) == i:
                    self.meta_data["matched"] = k
                    return self.meta_data["routes"][k]

    def _set_metadata(
            self,
            class_name: str,
            path_segment: str,
            var_type: Optional[str],
            var_name: Optional[str],
    ):
        """
        # Returns:
        #  {
        #     'route': '/images/image/{id:int}/pic/{file_name:str}',
        #     'split': ['images', 'image', '{id:int}', 'pic', '{file_name:str}'],
        #     'vars': {
        #         'id': {
        #             'name': 'id',
        #             'type': 'int',
        #             'value': '1',
        #         },
        #         'file_name': {
        #             'name': 'file_name',
        #             'type': 'str',
        #             'value': 'sunny.png',
        #         }
        #     },
        #     'path': '/images/image/1/pic/sunny.png',
        #     'class': 'Images'
        # }
        """
        if var_type is not None and var_name is not None:
            _route_vars = {
                f"{var_name}": {
                    "name": var_name,
                    "type": var_type,
                    "value": path_segment,
                }
            }
            if self.meta_data["routes"][class_name]["vars"]:
                _route_vars |= self.meta_data["routes"][class_name]["vars"]
            self.meta_data["routes"][class_name]["vars"] = _route_vars
        if "path" not in self.meta_data["routes"][class_name]:
            self.meta_data["routes"][class_name]["path"] = self.path
            self.meta_data["routes"][class_name]["class"] = class_name

    def route(self) -> Dict:
        """
        Matches the incoming path to a stored route handler path
        Important: Will match ONLY the first route handler class that matches the
        incoming request path.
        :return:
        :rtype:
        """
        route = self._match()
        return route

    def get_matched(self) -> Optional[str]:
        """
        :return:
        :rtype:
        """
        return self.meta_data["matched"]



from abc import ABC, abstractmethod
from typing import Tuple, Callable, TypeVar

from bobtail.response import Response
from bobtail.request import Request


class AbstractRoute(ABC):

    @abstractmethod
    def get(self, req: Request, res: Response) -> None:
        """
        :param req:
        :type req:
        :param res:
        :type res:
        :return:
        :rtype:
        """

    @abstractmethod
    def post(self, req: Request, res: Response) -> None:
        """
        :param req:
        :type req:
        :param res:
        :type res:
        :return:
        :rtype:
        """

    @abstractmethod
    def put(self, req: Request, res: Response) -> None:
        """
        :param req:
        :type req:
        :param res:
        :type res:
        :return:
        :rtype:
        """

    @abstractmethod
    def delete(self, req: Request, res: Response) -> None:
        """
        :param req:
        :type req:
        :param res:
        :type res:
        :return:
        :rtype:
        """


Handler = TypeVar("Handler", Callable[[Request, Response], None], None)
Route = TypeVar("Route", Tuple[AbstractRoute, str], None)



from abc import ABC


class BaseHeaders(ABC):

    content_type: str

    def __init__(self, content_type):
        self.content_type = content_type


class RequestHeaders(BaseHeaders):

    def __init__(self, content_type):
        super().__init__(content_type)

    def _get_type(self) -> str:
        return self.content_type.split(";")[0]

    def is_multipart(self) -> bool:
        """Is this a multipart-formdata request header"""
        if self._get_type() == "multipart/form-data":
            return True
        return False

    def is_urlencoded(self) -> bool:
        """Is this a application/x-www-form-urlencoded request header"""
        if self._get_type() == "application/x-www-form-urlencoded":
            return True
        return False


class ResponseHeaders(BaseHeaders):

    def __init__(self, content_type):
        super().__init__(content_type)



from typing import Dict, Union
from abc import ABC

from bobtail.wsgi_input import WSGIInput
from bobtail.headers import RequestHeaders


class Request(ABC):

    path: str

    method: str

    vars: Dict

    args: Dict

    headers: RequestHeaders

    byte_data: input

    wsgi_input: WSGIInput

    def __init__(self, *, path: str, method: str, byte_data: bytes, headers: RequestHeaders):
        self.path = path
        self.method = method
        self.headers = headers
        self.wsgi_input = WSGIInput(
            byte_data=byte_data,
            headers=self.headers,
        )

    def get_arg(self, name: str) -> Union[str, int, bool]:
        """
        :param name:
        :type name:
        :return:
        :rtype:
        """
        arg_value = self.args[name]["value"]
        match self.args[name]["type"]:
            case "int":
                return int(arg_value)
            case "str":
                return arg_value
            case "bool":
                return bool(arg_value == "true" or arg_value == "True")

    def set_args(self, args):
        self.args = args

    def get_json(self) -> Dict:
        """

        :return:
        :rtype:
        """
        return self.wsgi_input.get_json()

    def get_body(self) -> str:
        """
        Handles text/plain
        :return:
        :rtype:
        """
        return self.wsgi_input.get_body()

    def get_form_data(self) -> Dict:
        """
        Handles application/x-www-form-urlencoded
        :return:
        :rtype:
        """
        return self.wsgi_input.get_form_data()

    def get_multipart_data(self) -> Dict:
        """
        Handles multipart/form-data
        :return:
        :rtype:
        """
        return self.wsgi_input.get_multipart_data()



from typing import List, Callable
from abc import ABC, abstractmethod

from bobtail.route import Handler
from bobtail.response import Response
from bobtail.request import Request

Tail = Callable[[Request, Response], None]


class AbstractMiddleware(ABC):

    @abstractmethod
    def run(self, req: Request, res: Response, tail: Tail) -> None:
        pass


class Middleware:

    middlewares: List[AbstractMiddleware] = None

    handler: Handler = None

    request: Request

    response: Response

    def add(self, instance: AbstractMiddleware):
        if self.middlewares is None:
            self.middlewares = []
        self.middlewares.append(instance)

    def call(self, req: Request, res: Response, handler: Handler):
        if self.middlewares:
            self.handler = handler
            self.request = req
            self.response = res
            for middleware in self.middlewares:
                middleware.run(self.request, self.response, self.tail)
            self.handler(self.request, self.response)
            return
        handler(req, res)

    def tail(self, req: Request, res: Response) -> None:
        """
        :param req:
        :type req:
        :param res:
        :type res:
        :return:
        :rtype:
        """
        self.request = req
        self.response = res



from typing import List, Tuple, Dict

from bobtail.response import Response
from bobtail.request import Request
from bobtail.status import Status
from bobtail.exceptions import NoRoutesError, RouteClassError
from bobtail.route import Route, Handler
from bobtail.parser import Parser
from bobtail.middleware import Middleware
from bobtail.headers import ResponseHeaders, RequestHeaders
from bobtail.wsgi_input import WSGIInput


class BobTail:

    environ: Dict

    routes: List[Route]

    response: Response = None

    request: Request

    _status: Status

    _body: str

    parse_metadata: Dict = None

    middleware: Middleware = None

    request_headers: RequestHeaders

    response_headers: ResponseHeaders

    def _handle_404(self, req: Request, res: Response):
        self.response.set_status(404)

    def __init__(self, *args, **kwargs):
        if "routes" not in kwargs:
            raise NoRoutesError("Expected a list of routes")
        self.routes = kwargs["routes"]
        self.middleware = Middleware()

    def init_response(self):
        self.response = Response()

    def set_request(self):
        self.request = Request(
            path=self.environ["PATH_INFO"],
            method=self.environ["REQUEST_METHOD"],
            byte_data=self.environ["wsgi.input"].read(),
            headers=RequestHeaders(content_type=self.environ["CONTENT_TYPE"]),
        )

    def _call_handler(self, route: callable, method: str):
        if hasattr(route,  method):
            try:
                handler: Handler = getattr(route, method)
                if not handler:
                    self.middleware.call(self.request, self.response, self._handle_404)
                    return
                self.middleware.call(self.request, self.response, handler)
            except Exception as exc:
                raise RouteClassError("route class is not instantiated") from exc
        else:
            self.middleware.call(self.request, self.response, self._handle_404)

    def _handle_route(self):
        p = Parser(self.routes, self.request.path)
        self.parse_metadata = p.route()
        # Set the args on the request object
        if self.parse_metadata and "vars" in self.parse_metadata:
            self.request.set_args(self.parse_metadata["vars"])
        for current_route in self.routes:
            route, _ = current_route
            if route.__class__.__name__ == p.get_matched():
                match self.request.method:
                    case "GET":
                        self._call_handler(route, "get")
                        return
                    case "POST":
                        self._call_handler(route, "post")
                        return
                    case "DELETE":
                        self._call_handler(route, "delete")
                        return
                    case "PUT":
                        self._call_handler(route, "put")
                        return
                    case "PATCH":
                        self._call_handler(route, "patch")
                        return
        self.middleware.call(self.request, self.response, self._handle_404)

    def __call__(self, environ, start_response):
        self.environ = environ

        # Set request & response
        self.set_request()
        self.init_response()
        # Call route handler with default response
        self._handle_route()

        self._status = Status()
        status = self._status.get(self.response.status)

        response_headers = [("Content-type", "application/json")]

        # Start response
        start_response(status, response_headers)
        # Process the final byte list & headers
        data = self.response._process()
        return data

    def use(self, middleware):
        self.middleware.add(middleware)



from enum import Enum
from typing import Union, Dict, Optional
import json
from birman import Decoder

from bobtail.headers import RequestHeaders


class WSGIInput:

    byte_data: bytes

    headers: RequestHeaders

    def __init__(self, byte_data: bytes, headers: RequestHeaders):
        self.byte_data = byte_data
        self.headers = headers

    def get_json(self) -> Dict:
        if self.headers.content_type == "application/json":
            if self.byte_data:
                return json.loads(self.byte_data)
        return {}

    def get_form_data(self) -> Dict:
        if self.byte_data and self.headers.is_urlencoded():
            d = Decoder(self.byte_data)
            form_dict = d.decode()
            return form_dict
        return {}

    def get_multipart_data(self) -> Dict:
        if self.byte_data and self.headers.is_multipart():
            d = Decoder(self.byte_data)
            form_dict = d.decode()
            return form_dict
        return {}

    def get_body(self) -> str:
        return str(self.byte_data, "utf-8")



from typing import Dict, Tuple, List, Optional, Self
import json

from bobtail.request import Request


class Response:
    status: int = 200

    body: Dict = None

    headers: Dict = {
        "Content-Type": "application/json",
    }

    _content_length: int = None

    def _process(self) -> List[bytes]:
        """Process final response"""
        if self.body is not None:
            resp_data = bytes(json.dumps(self.body, indent=2), "utf-8")
            self.set_content_len(resp_data)
            return [resp_data]
        self.set_content_len(None)
        return []

    def set_content_len(self, data: Optional[bytes]) -> None:
        """
        Warning: calling this method directly inside a route handler will have no effect
        as the content length is set dynamically after the execution of the route handler
        :param data:
        :type data:
        :return:
        :rtype:
        """
        if self._content_length is None:
            con_len = 0 if data is None else len(data)
            self.headers = {
                **self.headers,
                "Content-Length": f"{con_len}",
            }

    def set_headers(self, value: Dict) -> None:
        """

        :param value: Excepts a dict
        :type value:
        :return:
        :rtype:
        """
        self.headers = {
            **self.headers,
            **value,
        }

    def set_status(self, value: int) -> None:
        """
        :param value: Sets the response status
        :type value:
        :return:
        :rtype:
        """
        self.status = value

    def set_body(self, value: Dict) -> None:
        """
        Set the body of the request
        :param value:
        :type value:
        :return:
        :rtype:
        """
        self.body = value



from setuptools import setup


with open("README.md", "r") as fh:
    long_description = fh.read()


setup(
    name="bobtail",
    version="0.0.8",
    description="A little Python http framework",
    packages=["bobtail"],
    py_modules=["bobtail"],
    install_requires=[
    ],
    classifiers=[
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3.11",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/joegasewicz/bobtail",
    author="Joe Gasewicz",
    author_email="joegasewicz@gmail.com",
)



from bobtail.exceptions import InvalidStatusError


class Status:

    codes = {
        "100": "Continue",
        "101": "Switching Protocols",
        "200": "OK",
        "201": "Created",
        "202": "Accepted",
        "203": "Non-Authoritative Information",
        "204": "No Content",
        "205": "Reset Content",
        "206": "Partial Content",
        "300": "Multiple Choices",
        "301": "Moved Permanently",
        "302": "Found",
        "303": "See Other",
        "304": "Not Modified",
        "305": "Use Proxy",
        "307": "Temporary Redirect",
        "400": "Bad Request",
        "401": "Unauthorized",
        "402": "Payment Required",
        "403": "Forbidden",
        "404": "Not Found",
        "405": "Method Not Allowed",
        "406": "Not Acceptable",
        "407": "Proxy Authentication Required",
        "408": "Request Timeout",
        "409": "Conflict",
        "410": "Gone",
        "411": "Length Required",
        "412": "Precondition Failed",
        "413": "Request Entity Too Large",
        "414": "Request-URI Too Long",
        "415": "Unsupported Media Type",
        "416": "Requested Range Not Satisfiable",
        "417": "Expectation Failed",
        "500": "Internal Server Error",
        "501": "Not Implemented",
        "502": "Bad Gateway",
        "503": "Service Unavailable",
        "504": "Gateway Timeout",
        "505": "HTTP Version Not Supported",
    }

    def get(self, status: int):
        """
        :param status:
        :type status:
        :return:
        :rtype:
        """
        try:
            return f"{status} {self.codes[str(status)]}"
        except KeyError as exc:
            raise InvalidStatusError("Invalid status") from exc
